import logging
import signal
import multiprocessing

from django.core.management.base import BaseCommand

from carrot.services import Worker
from carrot.settings import carrot_settings


LOGGER = logging.getLogger(__name__)


class Command(BaseCommand):
	def handle(self, *args, **options):
		LOGGER.info("Carrot starting")

		signal.signal(signal.SIGTERM, self._sigterm)
		signal.signal(signal.SIGINT, self._sigterm)

		workers = []
		for queue, settings in carrot_settings['queues'].items():
			for _ in range(int(settings['worker_concurrency'])):
				workers.append(Worker(queue))

		processes = []
		for worker in workers:
			p = multiprocessing.Process(target=worker.run, daemon=True)
			p.start()
			processes.append(p)

		for p in processes:
			p.join()

		LOGGER.info("Carrot shutdown complete")

	def _sigterm(self, signum, frame):
		# Nothing needs to be done. The forked worker processes will handle their own signal
		# and eventually the processes will be joined (hopefully)
		LOGGER.info(f"Carrot received shutdown signal ({signum}). Waiting for children to join")



import logging
from django.apps import AppConfig

LOGGER = logging.getLogger(__name__)


class CarrotConfig(AppConfig):
	name = 'carrot'
	verbose_name = "Carrots are better than celery"

	def ready(self):
		from carrot.connections import publisher  # noqa
		from carrot.settings import carrot_settings  # noqa

		# Initialize queues, exchanges
		publisher.connect()
		for queue, queue_settings in carrot_settings['queues'].items():
			LOGGER.info(f"Initializing queue ({queue_settings['queue_name']})")
			publisher.setup_queue_exchange(
				exchange=carrot_settings['exchange'],
				queue=queue_settings['queue_name'],
				routing_key=queue_settings['queue_name'],
				durable_exchange=carrot_settings['durable_exchange'],
				durable_queue=queue_settings['durable_queue'],
			)
		publisher.close()



import time
import logging
import signal

import pika

LOGGER = logging.getLogger(__name__)


class RabbitConnection:
	DEFAULT_EXCHANGE = ''
	DEFAULT_QUEUE = ''
	DEFAULT_ROUTING_KEY = ''
	DEFAULT_USER = 'guest'
	DEFAULT_PASSWORD = 'guest'
	DEFAULT_HOST = 'localhost'
	DEFAULT_PORT = '5672'
	DEFAULT_HEARTBEAT = 600
	DEFAULT_BLOCKED_CONNECTION_TIMEOUT = 300
	DEFAULT_MESSAGE_ENCODING = 'utf-8'
	DEFAULT_EXCHANGE_TYPE = 'direct'

	def __init__(
			self,
			user=DEFAULT_USER,
			password=DEFAULT_PASSWORD,
			host=DEFAULT_HOST,
			port=DEFAULT_PORT,
			queue=DEFAULT_QUEUE,
			exchange=DEFAULT_EXCHANGE,
			routing_key=DEFAULT_ROUTING_KEY,
			heartbeat=DEFAULT_HEARTBEAT,
			blocked_connection_timeout=DEFAULT_BLOCKED_CONNECTION_TIMEOUT,
			encoding=DEFAULT_MESSAGE_ENCODING,
	):
		# These must be able to cast as ints
		heartbeat = int(heartbeat)
		blocked_connection_timeout = int(blocked_connection_timeout)

		# The encoding for a message to use if no other is selected
		self.message_encoding = encoding

		# These aren't necessary during construction, but are useful for instance methods
		# If they aren't set here, those methods will expect their values passed in.
		self.queue = queue
		self.exchange = exchange
		self.routing_key = routing_key

		self._credentials = pika.PlainCredentials(username=user, password=password)
		self._parameters = pika.ConnectionParameters(**{
			'credentials': self._credentials,
			'host': host,
			'port': port,
			'heartbeat': heartbeat,
			'blocked_connection_timeout': blocked_connection_timeout,
		})

		self._connection = None
		self._channel = None

	def __enter__(self):
		self.connect()
		return self

	def __exit__(self, *args, **kwargs):
		self.close()

	@property
	def is_open(self):
		"""
		Test if channel is usable
		"""
		return self._channel is not None and self._channel.is_open

	def connect(self, parameters=None):
		"""
		Makes a connection to RabbitMQ server, and setups the the exchange and queue
		if it hasn't already been done.

		:param parameters: Optionally pass in ConnectionParameters if you want to override the defaults
		"""
		assert parameters is None or isinstance(parameters, pika.ConnectionParameters)
		parameters = parameters or self._parameters

		self._connection = pika.BlockingConnection(parameters)
		self._channel = self._connection.channel()

	def close(self):
		"""
		Close the RabbitMQ connection
		"""
		if self._connection is None:
			return

		try:
			if self._connection.is_open:
				self._connection.close()
		except (pika.exceptions.AMQPChannelError, pika.exceptions.AMQPConnectionError) as e:
			LOGGER.warning(f'Caught error while trying to close connection: {str(e)}')
			pass
		finally:
			self._connection = None
			self._channel = None

	def setup_queue_exchange(self, exchange=None, queue=None, routing_key=None, durable_queue=False, durable_exchange=False, exchange_type=DEFAULT_EXCHANGE_TYPE):
		"""
		Initialize the exchange and queue
		"""
		if not self.is_open:
			self.connect()

		exchange = exchange or self.exchange
		queue = queue or self.queue
		routing_key = routing_key or self.routing_key

		assert exchange, "You must provide a name for the exchange"
		assert queue, "You must provide a name for the queue"

		self._channel.exchange_declare(exchange=exchange, exchange_type=exchange_type, durable=durable_exchange)
		self._channel.queue_declare(queue, durable=durable_queue)
		self._channel.queue_bind(exchange=exchange, queue=queue, routing_key=routing_key)


class RabbitPublisher(RabbitConnection):
	"""
	A generic RabbitMQ publisher which can publish to any exchange

	Example:
	publisher = RabbitPublisher(exchange='task_queue')
	publisher.publish('loral igloo dolar')
	"""

	def publish(self, message, exchange=None, routing_key=None, encoding=None):
		"""
		Publish a message to rabbitmq. This is the main purpose of this class and it's main method
		If an attempt to publish fails because of a connection error, this will make one attempt
		to reconnect (which might be necessary due to a heartbeat timeout)

		:param str message: the body of the message to send
		"""
		assert isinstance(message, (str, bytes)), "The message must be a str or bytes object"
		if not self.is_open:
			self.connect()

		# This allows method kwarg overrides of self attrs (potentially publishing to another exchange)
		exchange = exchange or self.exchange
		routing_key = routing_key or self.routing_key

		# Exchange name is required, routing_key can be empty str
		assert exchange, "You must provide a name for the exchange"

		encoding = encoding or self.message_encoding
		properties = pika.BasicProperties(content_encoding=encoding)
		try:
			self._channel.basic_publish(exchange, routing_key, message, properties=properties)
		except pika.exceptions.AMQPConnectionError:
			LOGGER.warning(f'Connection was closed while publishing. Reconnecting...')
			self.connect()
			self._channel.basic_publish(exchange, routing_key, message, properties=properties)


class RabbitConsumer(RabbitConnection):
	"""
	A generic RabbitMQ consumer which gracefully handles SIGINTS and SIGTERMS

	Example:
	def print_article(x):
		print(x)
	consumer = RabbitConsumer(queue='newspaper', callback=print_article)
	consumer.run()
	...
	"""
	DEFAULT_RECONNECT_WAIT = 5
	DEFAULT_PREFETCH_COUNT = 2

	def __init__(self, callback=None, queue=None, prefetch_count=DEFAULT_PREFETCH_COUNT, *args, **kwargs):
		"""
		:param callback: callable function with w/ a single argument (the message body of a consumed message)
		:param queue: name of the queue to consume messages from
		:param prefetch_count: quantity of non-acked messages that can be obtained at any time
		"""
		super().__init__(*args, **kwargs)

		self.queue = queue or self.queue
		self.callback = callback

		self._prefetch_count = prefetch_count
		self._shutdown_flag = False

	def run(self, reconnect_wait=DEFAULT_RECONNECT_WAIT):
		if self.callback is None or callable(self.callback) is False:
			raise RuntimeError("A callback function must be set before running the consumer")

		# Signals to stop running (SIGINT, SIGTERM)
		LOGGER.info("Registering Signal handlers for: SIGINT, SIGTERM")
		signal.signal(signal.SIGTERM, self._sigterm)
		signal.signal(signal.SIGINT, self._sigterm)

		LOGGER.info("Starting consuming")
		self._shutdown_flag = False
		while self._shutdown_flag is False:
			try:
				self.connect()
				self._channel.basic_qos(prefetch_count=self._prefetch_count)

				self._channel.basic_consume(self.queue, self._on_message)
				self._channel.start_consuming()
				self._connection.close()

			except pika.exceptions.AMQPConnectionError as e:
				if not self._shutdown_flag:
					LOGGER.warning(f'Connection was closed. Reconnecting... ({e})')
					time.sleep(reconnect_wait)
				continue

			except pika.exceptions.ConnectionClosedByBroker:
				if not self._shutdown_flag:
					LOGGER.warning("Connection closed by broker. Will attempt reconnect shortly..")
					time.sleep(reconnect_wait)
				continue

			except pika.exceptions.AMQPChannelError:
				LOGGER.exception("Caught channel error. Stopping...")
				break

	def stop(self):
		self._shutdown_flag = True
		self._channel.stop_consuming()

	def _on_message(self, channel, method_frame, header_frame, body):
		"""
		WARNING: This normally shouldn't be overridden.
		The purpose of the consumer is to: retrieve messages, deliver them to a callback, and acknowledge message delivery.
		It is not the responsiblity of the consumer to ensure that the execution of the callable exited sucessfully.
		Therefore, this method serves as a callback wrapper to ensure that the consumer never dies, and that message delivery
		is always acknowledged. You have been warned.
		"""
		if not self._shutdown_flag:
			try:
				if self.message_encoding is not None:
					body = body.decode(self.message_encoding)
				self.callback(body)
			except:  # noqa
				LOGGER.exception("RabbitConsumer does not die, but the callable it delivers messages to raised an exception")
			channel.basic_ack(delivery_tag=method_frame.delivery_tag)

	def _sigterm(self, signum, frame):
		"""
		Gracefully handle SIGTERM and SIGINT
		"""
		LOGGER.info('Caught Sigterm. Shutting down gracefully')
		self.stop()



from carrot.amqp import RabbitPublisher
from carrot.settings import carrot_settings

publisher = RabbitPublisher(
	host=carrot_settings['host'],
	port=carrot_settings['port'],
	user=carrot_settings['user'],
	password=carrot_settings['password'],
	exchange=carrot_settings['exchange'],
)


def close_all():
	publisher.close()



# Generated by Django 3.0 on 2019-12-26 21:17

import carrot.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('kallable', models.CharField(max_length=512)),
                ('args', carrot.fields.ListField(blank=True, default='')),
                ('kwargs', carrot.fields.DictField(blank=True, default='')),
                ('queue', models.CharField(blank=True, default='', max_length=255)),
                ('status', models.SmallIntegerField(choices=[(1, 'Pending'), (2, 'Running'), (3, 'Completed'), (4, 'Failed')], db_index=True, default=1)),
                ('message', models.CharField(default='', max_length=2048)),
                ('exit_code', models.SmallIntegerField(choices=[(-1, 'Notset'), (0, 'Success'), (1, 'Unknown Error')], default=-1)),
                ('created_by', models.CharField(blank=True, default='', max_length=512)),
                ('created_on', models.DateTimeField(auto_now_add=True)),
                ('started_on', models.DateTimeField(blank=True, null=True)),
                ('completed_on', models.DateTimeField(blank=True, null=True)),
            ],
        ),
    ]



from django.db import models
from django.core.exceptions import ValidationError

import json


class ListField(models.TextField):
	"""
	ListField is a textfield that serializes/unserializes Lists
	It saves lists as JSON strings.
	"""
	def from_db_value(self, value, expression, connection):
		value = "[" + value + "]"
		return json.loads(value)

	def get_db_prep_save(self, value, *args, **kwargs):
		if value in ['', None]:
			value = []
		elif not isinstance(value, list):
			raise ValidationError("Invalid input for ListField: must be a list but got: %s" % type(value))

		value = json.dumps(value)
		value = value.lstrip('[').rstrip(']')
		return super().get_db_prep_save(value, *args, **kwargs)


class DictField(models.TextField):
	"""
	DictField is a textfield that serializes/unserializes Dict
	It saves lists as JSON strings.
	"""
	def from_db_value(self, value, expression, connection):
		value = "{" + value + "}"
		return json.loads(value)

	def get_db_prep_save(self, value, *args, **kwargs):
		if value in ['', None]:
			value = {}
		elif not isinstance(value, dict):
			raise ValidationError("Invalid input for DictField: must be a list but got: %s" % type(value))

		value = json.dumps(value)
		value = value.lstrip('{').rstrip('}')
		return super().get_db_prep_save(value, *args, **kwargs)



import logging

from django.db import models
from django.core.validators import ValidationError
from django.utils import timezone

from carrot.fields import ListField, DictField
from carrot.connections import publisher
from carrot.settings import carrot_settings, DEFAULT_QUEUE_NAME
from carrot.utils import import_callable


LOGGER = logging.getLogger(__name__)
EMPTY_STRING = ''
MESSAGE_FIELD_MAX_LEN = 2048


class Task(models.Model):
	"""
	Instances of Task represent an unexecuted call to a function with pre-selected args and kwargs.
	Task instances can publish themselves to a queue for asynchronous consumption, or can be executed
	directly with self.execute()
	"""
	class Status(models.TextChoices):
		PENDING = 'pending'
		RUNNING = 'running'
		COMPLETED = 'completed'
		FAILED = 'failed'

	class ExitCode(models.IntegerChoices):
		SUCCESS = 0
		UNKNOWN_ERROR = 1

	kallable = models.CharField(max_length=512)
	args = ListField(blank=True, default=EMPTY_STRING)
	kwargs = DictField(blank=True, default=EMPTY_STRING)

	queue = models.CharField(max_length=255, blank=True, default=DEFAULT_QUEUE_NAME)

	status = models.CharField(max_length=255, choices=Status.choices, default=Status.PENDING)
	message = models.CharField(max_length=MESSAGE_FIELD_MAX_LEN, default=EMPTY_STRING)
	exit_code = models.SmallIntegerField(choices=ExitCode.choices, blank=True, null=True)
	created_by = models.CharField(max_length=512, blank=True, default=EMPTY_STRING)

	created_on = models.DateTimeField(auto_now_add=True)
	started_on = models.DateTimeField(null=True, blank=True)
	completed_on = models.DateTimeField(null=True, blank=True)

	@property
	def can_publish(self):
		return self.queue in carrot_settings['queues'] and self.status == self.Status.PENDING

	def execute(self):
		"""
		Execute the function that this task represents
		"""
		# List/Dictfields may not be lists/dicts if the model instance
		# wasn't initialized from the db
		if not self.args:
			self.args = []
		if not self.kwargs:
			self.kwargs = {}

		self.started_on = timezone.now()
		self.status = self.Status.RUNNING
		if self.id:
			self.save(validate=False, update_fields={'status', 'started_on'})

		LOGGER.info(f"Executing task ({self.id}) => ({self.kallable})")
		try:
			self.exit_code = self.ExitCode.SUCCESS
			self.message = ''
			func = import_callable(self.kallable)
			return func(*self.args, **self.kwargs)

		except Exception as e:
			self.exit_code = self.ExitCode.UNKNOWN_ERROR
			self.message = repr(e)
			raise

		finally:
			self.status = self.Status.COMPLETED if self.exit_code == self.ExitCode.SUCCESS else self.Status.FAILED
			self.completed_on = timezone.now()
			if self.id:
				self.save(validate=False, update_fields={'status', 'exit_code', 'completed_on', 'message'})

	def publish(self):
		"""
		Publishes message to configured RabbitMQ connection if self.queue is a valid Queue member
		"""
		assert self.can_publish
		publisher.publish(
			message=str(self.id),
			exchange=carrot_settings['exchange'],
			routing_key=carrot_settings['queues'][self.queue]['queue_name'],
		)

	def validate(self):
		"""
		Used to verify that a kallable is both importable and is callable
		Also, used to verify that a queue is valid, if set
		"""
		# Make sure self.kallable is importable and is a callable object
		func = import_callable(self.kallable)

		if not callable(func):
			raise ValidationError(f"Callable ({self.kallable}) is not callable")

		# If instance has `queue` set, require that it exists
		if self.queue and self.queue not in carrot_settings['queues']:
			raise ValidationError(f"({self.queue}) is not a valid carrot queue")

	def save(self, *args, **kwargs):
		"""
		Enforces validation and publishes to queue during creation if queue is set
		"""
		if kwargs.pop('validate', True):
			self.validate()

		# Truncate the message if it's too long
		_message = str(self.message)
		if len(_message) > MESSAGE_FIELD_MAX_LEN:
			self.message = _message[:MESSAGE_FIELD_MAX_LEN]

		is_new = self._state.adding

		super().save(*args, **kwargs)
		if is_new and self.can_publish:
			self.publish()



import logging

from django.db import connections
from django.db.utils import OperationalError
from django.db.models import Q

from carrot.models import Task
from carrot.amqp import RabbitConsumer
from carrot.settings import carrot_settings


LOGGER = logging.getLogger(__name__)


class Worker(RabbitConsumer):
	"""
	Worker is a serial queued task consumer
	"""
	PREFETCH_COUNT = 1

	def __init__(self, queue, *args, **kwargs):
		assert queue in carrot_settings['queues'], f"{queue} is not a valid queue"

		kwargs['queue'] = carrot_settings['queues'][queue]['queue_name']
		kwargs['exchange'] = carrot_settings['exchange']
		kwargs['host'] = carrot_settings['host']
		kwargs['port'] = carrot_settings['port']
		kwargs['user'] = carrot_settings['user']
		kwargs['password'] = carrot_settings['password']

		kwargs['prefetch_count'] = self.PREFETCH_COUNT
		kwargs['callback'] = self.on_message

		super().__init__(*args, **kwargs)

	def run(self):
		# Ensure new db fds are opened since workers are commonly forks
		connections.close_all()
		return super().run()

	def on_message(self, message):
		try:
			task = self._get_task(message)
		except OperationalError:
			# client timeout might cause db connections to close. Attempt reconnect
			connections.close_all()
			task = self._get_task(message)

		if task is None:
			LOGGER.error(f"task ({message}) does not exist, was db_table flushed?")

		elif task.status != task.Status.PENDING:
			LOGGER.info(f"task ({message}) no longer pending execution, discarding.")

		else:
			try:
				task.execute()
			except: # noqa
				LOGGER.exception("Carrot Worker caught an exception, the task failed, but the worker does not die")

	def _get_task(self, task_id):
		try:
			task = Task.objects.get(id=task_id)
		except Task.DoesNotExist:
			task = None
		return task



import importlib


def import_callable(path):
	"""
	:param path: the dot-notated absolute path to a callable
	:returns: callable, e.g. function, class
	"""
	assert isinstance(path, str)
	split_path = path.split('.')
	module_path = ".".join(split_path[:-1])
	callable_name = split_path[-1]
	module = importlib.import_module(module_path)
	return getattr(module, callable_name)



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

# At the time of writing, the queue name was limited to 255 bytes of UTF-8 chars.
# https://www.rabbitmq.com/queues.html
MAX_BYTES_QUEUE_NAME = 255
DEFAULT_QUEUE_NAME = 'default'

DEFAULTS = {
	'host': '127.0.0.1',
	'port': '5672',
	'user': 'guest',
	'password': 'guest',
	'exchange': 'carrot.direct',
	'queue_prefix': 'carrot',
	'worker_concurrency': 1,
	'durable_queue': True,
	'durable_exchange': True,
	'durable_messages': False,
	'include_default_queue': True,
	'queues': {},
}

# Merge defaults with APP settings
carrot_settings = getattr(settings, 'CARROT', {})
assert isinstance(carrot_settings, dict), f'wrong type ({type(carrot_settings)}) for CARROT settings'
for k, v in DEFAULTS.items():
	carrot_settings.setdefault(k, v)

# Default queue
if carrot_settings['include_default_queue'] is True:
	carrot_settings['queues'].setdefault(DEFAULT_QUEUE_NAME, {})

# Setup and verify queues and their options
for queue, queue_settings in carrot_settings['queues'].items():
	assert isinstance(queue, str)
	assert isinstance(queue_settings, dict)

	# Set some defaults before validating
	queue_settings.setdefault('queue_name', f"{carrot_settings['queue_prefix']}.{queue}")
	queue_settings.setdefault('worker_concurrency', carrot_settings['worker_concurrency'])
	queue_settings.setdefault('durable_queue', carrot_settings['durable_queue'])

	# Validations
	queue_settings['worker_concurrency'] = int(queue_settings['worker_concurrency'])
	if not len(queue_settings['queue_name'].encode('utf-8')) <= MAX_BYTES_QUEUE_NAME:
		raise ImproperlyConfigured(f"Queue names may be up to {MAX_BYTES_QUEUE_NAME} bytes of UTF-8 characters")

assert isinstance(carrot_settings['host'], str)
assert isinstance(carrot_settings['port'], (int, str))
assert isinstance(carrot_settings['user'], str)
assert isinstance(carrot_settings['user'], str)
assert isinstance(carrot_settings['password'], str)
assert isinstance(carrot_settings['exchange'], str)
assert isinstance(carrot_settings['queue_prefix'], str)
assert isinstance(carrot_settings['worker_concurrency'], int)
assert isinstance(carrot_settings['durable_queue'], bool)
assert isinstance(carrot_settings['durable_exchange'], bool)
assert isinstance(carrot_settings['durable_messages'], bool)
assert isinstance(carrot_settings['include_default_queue'], bool)
assert isinstance(carrot_settings['queues'], dict)



#!/usr/bin/env python3

import os
from setuptools import setup, find_packages

BASE_DIR = os.path.dirname(__file__)
VERSION = '2.3.1'

setup(
  name = 'djangocarrot',
  version = VERSION,
  description = 'Carrot -- A simple task Queue for Django.',
  author = 'John Bowen',
  author_email = 'jbowen7@gmail.com',
  url = 'https://github.com/jbowen7/djangocarrot',
  keywords = ['django', 'task', 'queue'],
  install_requires = ['pika',],
  package_dir={'': 'src'},
  packages=find_packages(where='src', exclude=[]),
)



# Generated by Django 3.0 on 2020-03-24 23:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('carrot', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='task',
            name='exit_code',
            field=models.SmallIntegerField(blank=True, choices=[(0, 'Success'), (1, 'Unknown Error')], null=True),
        ),
        migrations.AlterField(
            model_name='task',
            name='queue',
            field=models.CharField(blank=True, default='default', max_length=255),
        ),
        migrations.AlterField(
            model_name='task',
            name='status',
            field=models.CharField(choices=[('pending', 'Pending'), ('running', 'Running'), ('completed', 'Completed'), ('failed', 'Failed')], default='pending', max_length=255),
        ),
    ]
